from simulator.environment import Env
import pygame
from simulator import globals
from simulator.utils import get_lane_borders
from simulator.traffic_router import TrafficRouter
from simulator.host_vehicle import HostVehicle
import simulator
import math
import random
from simulator.traffic import SingleTargetAhead

pygame.font.init()


class Simulator:
    def __init__(self, number_of_lanes, speed_limit, initial_speed=27, show_parameter_panel=True, show_odometer_panel=True, size="medium"):
        globals.set_screen_size(size=size)
        self.screen = pygame.display.set_mode((globals.WIDTH,globals.HEIGHT))

        pygame.display.set_caption(globals.CAPTION)
        clock = pygame.time.Clock()
        self.num_lanes = number_of_lanes
        # create sprites
        self.env = Env(number_of_lanes, show_horizontal_line=False)
        self.road_borders = self.env.get_sidwalk_road_borders()
        self.lane_borders = get_lane_borders(self.road_borders[0], num_lane=self.num_lanes)
        self.show_parameter_panel = show_parameter_panel

        # ----------------------------------- SETUP TRAFFIC ---------------------------------
        # speed range of traffic objects
        low_speed = math.floor(0.7 * speed_limit)
        high_speed = math.floor(0.9 * speed_limit)

        # router class needs below parameters
        # gap_offset: offset for gap between two adjacent traffic objects (Y axis)
        # traffic_object_start_point: start point for adding new traffic objects (Y axis)
        # has_speed_change: should change the speed of traffic objects or not
        # speed_change_gap_low and speed_change_gap_high: range for spent time to change the speed of traffic objects
        self.traffic_router = TrafficRouter(self.lane_borders, self.road_borders, self.num_lanes,
                                            speed_range=[low_speed, high_speed], gap_offset=0,
                                            traffic_object_start_point=0, has_speed_change=True,
                                            new_traffic_object_probability=0.33)

        # -------------------------------- SETUP HOST VEHICLE ---------------------------------
        self.host_vehicle = HostVehicle(globals.HOST_VEHICLE_IMAGES, self.lane_borders,
                                        p_bottom_pos=(0,globals.HEIGHT),
                                        p_init_lng_spd=initial_speed, p_max_spd=speed_limit, is_motorcycle=False, lane=1)
        self.host_odometer_value = 0
        self.show_odometer_panel = show_parameter_panel
        # initialize scenario variables
        self.running = True
        self.perception = {}
        self.lanes_speed = [0] * self.num_lanes
        self.clock = pygame.time.Clock()
        self.quit_game = False

    def step(self, action="", user_given_odometer_value=0):

        if self.running:

            # update host vehicle based on user input
            self.host_vehicle.update(action)

            # update the environment
            self.env.update(speed=self.host_vehicle.get_speed())

            # manage traffic objects (we should provide host vehicle lane in each step to manage the speed of that lane by specified step)
            targets, self.lanes_speed = self.traffic_router.update(host_speed=self.host_vehicle.get_speed())

            # get the self.perception from host vehicle
            self.perception = self.host_vehicle.sensing(targets)

            # update odometer value
            self.host_odometer_value += self.host_vehicle.get_speed()

            # check the possible collision
            self.running = self.perception.running

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.running = False
                    self.quit_game = True
                    self.perception.running = False
                    pygame.quit()
                    exit()

            self.render(user_given_odometer_value=user_given_odometer_value)
            return self.perception
        else:
            self.render(user_given_odometer_value=user_given_odometer_value)
            return {"running":False}

    def render(self, user_given_odometer_value=0):
        if not self.quit_game:
            if self.running:
                self.env.render(self.screen)

                # render the traffic objects
                self.traffic_router.render(self.screen)

                # render host vehicle
                self.host_vehicle.render(self.screen)

                if self.show_parameter_panel:

                    data = [
                        f"host_speed: {self.perception.host_speed}",
                        f"host_lane: {self.perception.host_lane}",
                        "",
                        f"forward dist: {self.perception.forward_target.dist}",
                        f"forward speed: {self.perception.forward_target.vy}",
                        "",
                        f"left_lane_available: {self.perception.left_lane_available}",
                        f"left dist: {self.perception.left_forward_target.dist}",
                        f"left speed: {self.perception.left_forward_target.vy}",
                        "",
                        f"right_lane_available: {self.perception.right_lane_available}",
                        f"right dist: {self.perception.right_forward_target.dist}",
                        f"right speed: {self.perception.right_forward_target.vy}",
                        "",
                        f"lanes speed: {self.lanes_speed}"
                    ]

                pygame.draw.rect(self.screen, globals.YELLOW, pygame.Rect(globals.PARAMETER_PANEL_YELLOW_RECT_START_X, globals.PARAMETER_PANEL_YELLOW_RECT_START_y, globals.PARAMETER_PANEL_YELLOW_RECT_WIDTH, globals.PARAMETER_PANEL_YELLOW_RECT_HEIGHT))
                pygame.draw.rect(self.screen, globals.BLACK, pygame.Rect(globals.PARAMETER_PANEL_BLACK_RECT_START_X, globals.PARAMETER_PANEL_BLACK_RECT_START_y, globals.PARAMETER_PANEL_BLACK_RECT_WIDTH, globals.PARAMETER_PANEL_BLACK_RECT_HEIGHT))
                for idx in range(len(data)):
                    font = pygame.font.Font('freesansbold.ttf', globals.PANEL_FONT_SIZE)
                    text = font.render(data[idx], True, globals.WHITE)
                    self.screen.blit(text, (globals.PARAMETER_PANEL_TEXT_START_X, globals.PARAMETER_PANEL_BLACK_RECT_START_y + globals.PADDING + idx * globals.PARAMETER_PANEL_TEXT_HEIGHT))

            if self.show_odometer_panel:
                pygame.draw.rect(self.screen, globals.YELLOW, pygame.Rect(globals.ODOMETER_PANEL_START_POINT, globals.ODOMETER_PANEL_START_POINT, globals.ODOMETER_PANEL_WIDTH + 2 * globals.ODOMETER_PANEL_BORDER_WIDTH, globals.ODOMETER_PANEL_HEIGHT + 2 * globals.ODOMETER_PANEL_BORDER_WIDTH))
                pygame.draw.rect(self.screen, globals.BLACK, pygame.Rect(globals.ODOMETER_PANEL_START_POINT + globals.ODOMETER_PANEL_BORDER_WIDTH, globals.ODOMETER_PANEL_START_POINT + globals.ODOMETER_PANEL_BORDER_WIDTH, globals.ODOMETER_PANEL_WIDTH, globals.ODOMETER_PANEL_HEIGHT))
                font = pygame.font.Font('freesansbold.ttf', globals.PANEL_FONT_SIZE)
                text = font.render("ODOMETER PANEL", True, globals.YELLOW)
                self.screen.blit(text, (globals.ODOMETER_PANEL_START_POINT * 2, globals.ODOMETER_PANEL_START_POINT * 2))
                font = pygame.font.Font('freesansbold.ttf', globals.ODOMETER_FONT_SIZE)
                text = font.render(f"USER VALUE: {user_given_odometer_value:,}", True, globals.WHITE)
                self.screen.blit(text, (globals.ODOMETER_PANEL_START_POINT * 2, globals.ODOMETER_PANEL_START_POINT * 2 + globals.ODOMETER_PANEL_HEIGHT * 1 / 3))
                text = font.render(f"REAL VALUE: {self.host_odometer_value:,}", True, globals.WHITE)
                self.screen.blit(text, (globals.ODOMETER_PANEL_START_POINT * 2, globals.ODOMETER_PANEL_START_POINT * 2 + globals.ODOMETER_PANEL_HEIGHT * 2 / 3))

                pygame.display.flip()
            else:
                # draw/render
                self.env.render(self.screen)
                font = pygame.font.Font('freesansbold.ttf', globals.FONT_SIZE)
                text = font.render('SIMULATION ENDED', True, globals.GREEN, globals.BLUE)
                text_rect = text.get_rect(center=(globals.WIDTH / 2, globals.HEIGHT / 2))
                self.screen.blit(text, text_rect)
                pygame.display.flip()

            pygame.display.update()
            self.clock.tick(globals.FPS)


class MultiLaneHighwayAVSimulator:
    def __init__(self, speed_limit=30, initial_speed=27, show_parameter_panel=True, size="medium"):
        globals.set_screen_size(size=size)
        self.screen = pygame.display.set_mode((globals.WIDTH,globals.HEIGHT))

        pygame.display.set_caption(globals.CAPTION)
        clock = pygame.time.Clock()
        self.num_lanes = 3
        # create sprites
        self.env = Env(self.num_lanes, show_horizontal_line=False)
        self.road_borders = self.env.get_sidwalk_road_borders()
        self.lane_borders = get_lane_borders(self.road_borders[0], num_lane=self.num_lanes)
        self.show_parameter_panel = show_parameter_panel

        # ----------------------------------- SETUP TRAFFIC ---------------------------------
        # speed range of traffic objects
        low_speed = math.floor(0.7 * speed_limit)
        high_speed = math.floor(0.9 * speed_limit)

        # router class needs below parameters
        # gap_offset: offset for gap between two adjacent traffic objects (Y axis)
        # traffic_object_start_point: start point for adding new traffic objects (Y axis)
        # has_speed_change: should change the speed of traffic objects or not
        # speed_change_gap_low and speed_change_gap_high: range for spent time to change the speed of traffic objects
        self.traffic_router = TrafficRouter(self.lane_borders, self.road_borders, self.num_lanes,
                                            speed_range=[low_speed, high_speed], gap_offset=0,
                                            traffic_object_start_point=0, has_speed_change=True,
                                            new_traffic_object_probability=0.33)

        # -------------------------------- SETUP HOST VEHICLE ---------------------------------
        self.host_vehicle = HostVehicle(globals.HOST_VEHICLE_IMAGES, self.lane_borders,
                                        p_bottom_pos=(0,globals.HEIGHT),
                                        p_init_lng_spd=initial_speed, p_max_spd=speed_limit, is_motorcycle=False, lane=1)
        self.host_odometer_value = 0
        self.show_odometer_panel = show_parameter_panel
        # initialize scenario variables
        self.running = True
        self.perception = {}
        self.lanes_speed = [0] * self.num_lanes
        self.clock = pygame.time.Clock()
        self.quit_game = False

    def step(self, action=""):
        if self.running:

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.running = False
                    self.quit_game = True
                    self.perception.running = False
                    pygame.quit()
                    exit()
            # update host vehicle based on user input
            self.host_vehicle.update(action)

            # update the environment
            self.env.update(speed=self.host_vehicle.get_speed())

            # manage traffic objects (we should provide host vehicle lane in each step to manage the speed of that lane by specified step)
            targets, self.lanes_speed = self.traffic_router.update(host_speed=self.host_vehicle.get_speed())

            # get the self.perception from host vehicle
            self.perception = self.host_vehicle.sensing(targets)

            # update odometer value
            self.host_odometer_value += self.host_vehicle.get_speed()

            # check the possible collision
            self.running = self.perception.running

            self.render()
            return self.perception
        else:
            return {"running":False}

    def render(self):
        if self.running:
            self.env.render(self.screen)

            # render the traffic objects
            self.traffic_router.render(self.screen)

            # render host vehicle
            self.host_vehicle.render(self.screen)

            if self.show_parameter_panel:

                data = [
                    f"host_speed: {self.perception.host_speed}",
                    f"host_lane: {self.perception.host_lane}",
                    "",
                    f"forward dist: {self.perception.forward_target.dist}",
                    f"forward speed: {self.perception.forward_target.vy}",
                    "",
                    f"left_lane_available: {self.perception.left_lane_available}",
                    f"left dist: {self.perception.left_forward_target.dist}",
                    f"left speed: {self.perception.left_forward_target.vy}",
                    "",
                    f"right_lane_available: {self.perception.right_lane_available}",
                    f"right dist: {self.perception.right_forward_target.dist}",
                    f"right speed: {self.perception.right_forward_target.vy}",
                    "",
                    f"lanes speed: {self.lanes_speed}"
                ]

                pygame.draw.rect(self.screen, globals.YELLOW, pygame.Rect(globals.PARAMETER_PANEL_YELLOW_RECT_START_X, globals.PARAMETER_PANEL_YELLOW_RECT_START_y, globals.PARAMETER_PANEL_YELLOW_RECT_WIDTH, globals.PARAMETER_PANEL_YELLOW_RECT_HEIGHT))
                pygame.draw.rect(self.screen, globals.BLACK, pygame.Rect(globals.PARAMETER_PANEL_BLACK_RECT_START_X, globals.PARAMETER_PANEL_BLACK_RECT_START_y, globals.PARAMETER_PANEL_BLACK_RECT_WIDTH, globals.PARAMETER_PANEL_BLACK_RECT_HEIGHT))
                for idx in range(len(data)):
                    font = pygame.font.Font('freesansbold.ttf', globals.PANEL_FONT_SIZE)
                    text = font.render(data[idx], True, globals.WHITE)
                    self.screen.blit(text, (globals.PARAMETER_PANEL_TEXT_START_X, globals.PARAMETER_PANEL_BLACK_RECT_START_y + globals.PADDING + idx * globals.PARAMETER_PANEL_TEXT_HEIGHT))

            pygame.display.flip()
        else:
            # draw/render
            self.env.render(self.screen)
            font = pygame.font.Font('freesansbold.ttf', globals.FONT_SIZE)
            text = font.render('SIMULATION ENDED', True, globals.GREEN, globals.BLUE)
            text_rect = text.get_rect(center=(globals.WIDTH / 2, globals.HEIGHT / 2))
            self.screen.blit(text, text_rect)
            pygame.display.flip()

        pygame.display.update()
        self.clock.tick(globals.FPS)


class OdometerCalculationSimulator:
    def __init__(self, size="medium", initial_speed=25, odometer_threshold=500, speed_range=[5, 10], is_while_loop=False):
        self.is_while_loop = is_while_loop
        self.speed_range = speed_range
        globals.set_screen_size(size=size)
        self.screen = pygame.display.set_mode((globals.WIDTH,globals.HEIGHT))

        pygame.display.set_caption(globals.CAPTION)
        self.num_lanes = 1
        self.odometer_threshold = odometer_threshold
        # create sprites
        self.env = Env(self.num_lanes, show_horizontal_line=False)

        self.road_borders = self.env.get_sidwalk_road_borders()
        self.lane_borders = get_lane_borders(self.road_borders[0], num_lane=self.num_lanes)

        # -------------------------------- SETUP HOST VEHICLE ---------------------------------
        self.host_vehicle = HostVehicle(globals.HOST_VEHICLE_IMAGES, self.lane_borders,
                                        p_bottom_pos=(0,globals.HEIGHT),
                                        p_init_lng_spd=initial_speed, is_motorcycle=False, lane=0)
        self.host_odometer_value = 0
        self.user_given_odometer_value = 0
        # initialize scenario variables
        self.running = True
        self.clock = pygame.time.Clock()
        self.quit_game = False

        self.time_step = 1
        self.message = ""

    def step(self, action="", user_given_odometer_value=0):
        self.user_given_odometer_value = user_given_odometer_value

        if self.running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.running = False
                    self.quit_game = True
                    pygame.quit()
                    exit()

            if (self.host_odometer_value - self.odometer_threshold) >= self.host_vehicle.get_speed():
                self.message = "FAIL: TRAVELLED PAST THE REQUIRED DISTANCE!"
            elif (self.host_odometer_value - self.odometer_threshold) < 0 and action == "stop":
                self.message = "FAIL: DID NOT TRAVEL THE REQUIRED DISTANCE!"
            elif action == "stop":
                self.message = "PASS: TRAVELLED THE CORRECT DISTANCE!"

            self.render()
            if self.is_while_loop:
                new_speed = random.randint(self.speed_range[0], self.speed_range[1])
                self.host_vehicle.set_speed(new_speed)

            # update host vehicle based on user input
            if globals.DEBUG:
                print(f"speed outside env: {self.host_vehicle.get_speed()}")
            self.host_vehicle.update("")

            # update the environment
            self.env.update(speed=self.host_vehicle.get_speed())

            # update odometer value
            self.host_odometer_value += self.host_vehicle.get_speed()

            return self.host_vehicle.get_speed(), self.time_step

        else:
            self.render()
            return {"running":False}

    def render(self):
        if not self.quit_game:
            if self.running:
                self.env.render(self.screen)

                # render host vehicle
                self.host_vehicle.render(self.screen)

            pygame.draw.rect(self.screen, globals.YELLOW, pygame.Rect(globals.ODOMETER_PANEL_START_POINT, globals.ODOMETER_PANEL_START_POINT, globals.ODOMETER_PANEL_WIDTH + 2 * globals.ODOMETER_PANEL_BORDER_WIDTH, globals.ODOMETER_PANEL_HEIGHT + 2 * globals.ODOMETER_PANEL_BORDER_WIDTH))
            pygame.draw.rect(self.screen, globals.BLACK, pygame.Rect(globals.ODOMETER_PANEL_START_POINT + globals.ODOMETER_PANEL_BORDER_WIDTH, globals.ODOMETER_PANEL_START_POINT + globals.ODOMETER_PANEL_BORDER_WIDTH, globals.ODOMETER_PANEL_WIDTH, globals.ODOMETER_PANEL_HEIGHT))
            font = pygame.font.Font('freesansbold.ttf', globals.PANEL_FONT_SIZE)
            text = font.render("ODOMETER PANEL", True, globals.YELLOW)
            self.screen.blit(text, (globals.ODOMETER_PANEL_START_POINT * 2, globals.ODOMETER_PANEL_START_POINT * 2))
            font = pygame.font.Font('freesansbold.ttf', globals.ODOMETER_FONT_SIZE)
            color = globals.RED
            if int(self.user_given_odometer_value) == int(self.host_odometer_value):
                color = globals.GREEN

            text = font.render(f"USER VALUE: {self.user_given_odometer_value:,}", True, globals.WHITE)
            self.screen.blit(text, (globals.ODOMETER_PANEL_START_POINT * 2, globals.ODOMETER_PANEL_START_POINT * 2 + globals.ODOMETER_PANEL_HEIGHT * 1 / 4))
            delta = self.host_odometer_value - self.user_given_odometer_value
            text = font.render(f"ERROR: {delta:,}", True, color)
            self.screen.blit(text, (globals.ODOMETER_PANEL_START_POINT * 2, globals.ODOMETER_PANEL_START_POINT * 2 + globals.ODOMETER_PANEL_HEIGHT * 2 / 4))
            text = font.render(f"SPEED: {self.host_vehicle.get_speed()}", True, globals.WHITE)
            self.screen.blit(text, (globals.ODOMETER_PANEL_START_POINT * 2, globals.ODOMETER_PANEL_START_POINT * 2 + globals.ODOMETER_PANEL_HEIGHT * 3 / 4))

            if self.message != "":
                font = pygame.font.Font('freesansbold.ttf', globals.FONT_SIZE)
                if "PASS" in self.message:
                    text = font.render(self.message, True, globals.GREEN, globals.BLACK)
                else:
                    text = font.render(self.message, True, globals.RED, globals.BLACK)
                text_rect = text.get_rect(center=(globals.WIDTH / 2, globals.HEIGHT / 2))
                self.screen.blit(text, text_rect)

            pygame.display.flip()
        else:
            # draw/render
            self.env.render(self.screen)
            font = pygame.font.Font('freesansbold.ttf', globals.FONT_SIZE)
            text = font.render('SIMULATION ENDED', True, globals.GREEN, globals.BLUE)
            text_rect = text.get_rect(center=(globals.WIDTH / 2, globals.HEIGHT / 2))
            self.screen.blit(text, text_rect)
            pygame.display.flip()

            pygame.display.update()
            self.clock.tick(globals.FPS)


class ParallelParkingSimulator:
    def __init__(self, initial_speed=20, size="medium"):
        globals.set_screen_size(size=size)
        self.screen = pygame.display.set_mode((globals.WIDTH,globals.HEIGHT))

        pygame.display.set_caption(globals.CAPTION)
        self.num_lanes = 2
        # create sprites
        self.env = Env(self.num_lanes, show_horizontal_line=True)
        self.clock = pygame.time.Clock()

        self.check_position_list = []
        self.step_position_list = []
        self.step_image_list = []
        self.x_image = pygame.transform.scale(pygame.image.load(globals.X_IMAGE).convert_alpha(), globals.CHECK_SIZE)
        self.check_image = pygame.transform.scale(pygame.image.load(globals.CHECK_IMAGE).convert_alpha(), globals.CHECK_SIZE)
        for i in range(0, 7):
            self.step_image_list.append(pygame.transform.scale(pygame.image.load(globals.STEPS_IMAGES[i]).convert_alpha(), globals.STEP_SIZE))
            self.check_position_list.append(globals.PARKING_PANEL_START_POINT * 2 + i * globals.STEP_GAP)
            self.step_position_list.append(globals.PARKING_PANEL_START_POINT * 2 + i * globals.STEP_GAP)

        self.itr_step = 0
        self.required_action = "forward"
        self.failure_message = ""
        self.done_steps = [False, False, False, False, False, False, False]
        self.done_steps = [False, False, False, False, False, False, False]

        self.road_borders = self.env.get_sidwalk_road_borders()
        self.lane_borders = get_lane_borders(self.road_borders[0], num_lane=self.num_lanes)
        self.get_horizontal_line_location = self.env.get_horizontal_line_location()
        # add two globals.VEHICLES
        self.targets = pygame.sprite.Group()
        image_idx = random.randint(0,3)
        target_1 = SingleTargetAhead(globals.VEHICLE_IMAGES[image_idx], self.lane_borders,
                                     p_bottom_pos=(0,self.get_horizontal_line_location[1][1] - 30),
                                     p_init_lng_spd=0, is_motorcycle=False, lane=1)
        image_idx = random.randint(0,3)
        target_2 = SingleTargetAhead(globals.VEHICLE_IMAGES[image_idx], self.lane_borders,
                                     p_bottom_pos=(0,self.get_horizontal_line_location[2][1] + 30),
                                     p_init_lng_spd=0, is_motorcycle=False, lane=1, set_top=True)
        self.targets.add(target_1)
        self.targets.add(target_2)

        # -------------------------------- SETUP HOST VEHICLE ---------------------------------
        self.host_vehicle = HostVehicle(globals.HOST_VEHICLE_IMAGES, self.lane_borders,
                                        p_bottom_pos=(0,globals.HEIGHT),
                                        p_init_lng_spd=initial_speed, is_motorcycle=False, lane=0)

        # initialize scenario variables
        self.running = True
        self.clock = pygame.time.Clock()
        self.quit_game = False

    def step(self, action):
        if self.running and self.failure_message == "":
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.running = False
                    self.quit_game = True
                    pygame.quit()
                    exit()

            if action == self.required_action and self.failure_message == "":
                # update host vehicle based on user input
                self.host_vehicle.update(action)

                # update the environment
                self.env.update(speed=0)

                if self.itr_step == globals.PARALLEL_PARKING_STEPS[0]:
                    self.done_steps[0] = True
                    self.required_action = "turn_right"
                elif self.itr_step == globals.PARALLEL_PARKING_STEPS[1]:
                    self.done_steps[1] = True
                    self.required_action = "reverse"
                elif self.itr_step == globals.PARALLEL_PARKING_STEPS[2]:
                    self.done_steps[2] = True
                    self.required_action = "turn_left"
                elif self.itr_step == globals.PARALLEL_PARKING_STEPS[3]:
                    self.done_steps[3] = True
                    self.required_action = "reverse"
                elif self.itr_step == globals.PARALLEL_PARKING_STEPS[4]:
                    self.done_steps[4] = True
                    self.required_action = "turn_left"
                elif self.itr_step == globals.PARALLEL_PARKING_STEPS[5]:
                    self.done_steps[5] = True
                    self.required_action = "reverse"
                elif self.itr_step == globals.PARALLEL_PARKING_STEPS[6]:
                    self.done_steps[6] = True
                    self.required_action = "show_result"

                self.itr_step += 1
            elif self.required_action != "done":

                self.failure_message = f"You should use '{self.required_action}' action here"

            if globals.DEBUG:
                print(f"step: {self.itr_step} - action: {action} - next required_action: {self.required_action}")

            self.render()

        else:
            self.render()
            return {"running":False}

    def render(self):
        if not self.quit_game:
            if self.failure_message == "":
                self.env.render(self.screen, show_sidewalk_objects=False)

                self.targets.draw(self.screen)

                # render host vehicle
                self.host_vehicle.parrallel_parking_render(self.screen)

            if self.required_action == "show_result":
                font = pygame.font.Font('freesansbold.ttf', globals.FONT_SIZE)
                text = font.render("WELL DONE!", True, globals.GREEN, globals.BLUE)
                text_rect = text.get_rect(center=(globals.WIDTH / 2, globals.HEIGHT / 2))
                self.screen.blit(text, text_rect)
        else:
            self.env.render(self.screen)

            self.targets.draw(self.screen)

            # render host vehicle
            self.host_vehicle.parrallel_parking_render(self.screen)

            font = pygame.font.Font('freesansbold.ttf', globals.FONT_SIZE)
            text = font.render(self.failure_message, True, globals.GREEN, globals.BLUE)
            text_rect = text.get_rect(center=(globals.WIDTH / 2, globals.HEIGHT / 2))
            self.screen.blit(text, text_rect)

        for i in range(0, 7):
            self.screen.blit(self.step_image_list[i], (globals.PARKING_PANEL_START_POINT, self.step_position_list[i]))

        for i in range(7):
            if self.done_steps[i] == True:
                self.screen.blit(self.check_image, (globals.STEP_SIZE[0] - globals.CHECK_SIZE[0], self.check_position_list[i]))
            elif self.failure_message != "":
                self.screen.blit(self.x_image, (globals.STEP_SIZE[0] - globals.CHECK_SIZE[0], self.check_position_list[i]))
                break

        pygame.display.update()
        self.clock.tick(globals.FPS)



from simulator.vehicle import Vehicle


class HostVehicle(Vehicle):
    def __init__(self, p_image, lane_borders, lane=0, p_bottom_pos=(0,0), p_init_lat_spd=0, p_init_lng_spd=0, p_max_spd=30, p_cycle_limit=1000, is_motorcycle=False):
        super().__init__(p_image, lane_borders, p_bottom_pos, p_init_lat_spd, p_init_lng_spd, p_max_spd, is_motorcycle)

        # set target speed
        self.set_speed(self.vy)

        # set target lane
        self.go_to_lane(lane)

        # simulation attributes
        self.status = 1  # (1; active, -1: terminate)
        self.cycle = 0
        self.cycle_limit = p_cycle_limit

    def update(self, action):
        if action == "speed_up":
            self.speed_up()
        elif action == "speed_down":
            self.speed_down()
        elif action == "shift_left":
            self.move_left()
        elif action == "shift_right":
            self.move_right()
        elif action == "stop":
            self.stop()
        elif action == "turn_right":
            self.turn_right()
        elif action == "turn_left":
            self.turn_left()
        elif action == "forward":
            self.forward()
        elif action == "reverse":
            self.reverse()
        else:
            pass

    # renders the host vehicle on the display
    def render(self, screen):
        screen.blit(self.image, self.get_rect_position())



from simulator import globals

from easydict import EasyDict as edict


def calculate_roadside_area_width(remained_width, grass_area_ratio, sidewalk_area_ratio):
    grass_area_width = remained_width / (2 * (grass_area_ratio + sidewalk_area_ratio))
    sidewalk_area_width = grass_area_width * (sidewalk_area_ratio / grass_area_ratio)
    return edict({
        "grass_area_width": grass_area_width,
        "sidewalk_area_width": sidewalk_area_width
    })


def get_sidewalk_x_loc(screen_width, grass_area_width, grass_area_ratio, sidewalk_area_ratio, pos_mode, left_side=False):
    if left_side:
        return screen_width - grass_area_width * (grass_area_ratio + sidewalk_area_ratio * 0.75) + pos_mode * grass_area_width * (sidewalk_area_ratio * 0.5)
    else:
        return grass_area_width * (grass_area_ratio + sidewalk_area_ratio * 0.75) - pos_mode * grass_area_width * (sidewalk_area_ratio * 0.5)


def get_lane_borders(start_point, num_lane):
    return [(start_point + i * (globals.ROAD_LANE_WIDTH + globals.WHITE_LANE_WIDTH), start_point + (i + 1) * globals.ROAD_LANE_WIDTH + i * globals.WHITE_LANE_WIDTH) for i in range(0, num_lane)]


def get_lane(p_x_coord, lane_borders):
    result = -1
    lane_number = 0
    for lane in lane_borders:
        if lane[0] <= p_x_coord <= lane[1]:
            result = lane_number
            break

        lane_number += 1

    return result


def get_lane_coordinate(p_lane_num, lane_borders):
    result = -1
    if p_lane_num < len(lane_borders):
        temp = lane_borders[p_lane_num]
        result = (temp[0] + temp[1]) / 2

    return result


def is_lane_available(p_x_coord, lane_borders, side):
    if side == "left":
        return get_lane(p_x_coord, lane_borders) > 0
    elif side == "right":
        return get_lane(p_x_coord, lane_borders) < (len(lane_borders) - 1)


def is_on_screen(p_x_coord, p_y_coord):
    result = False
    if 0 <= p_x_coord <= globals.WIDTH and 0 <= p_y_coord <= globals.HEIGHT:
        result = True
    return result



import pygame
from simulator import globals
from simulator.utils import get_lane, get_lane_coordinate, is_on_screen, is_lane_available
from easydict import EasyDict as edict
import math


class Vehicle(pygame.sprite.Sprite):
    def __init__(self, p_image, lane_borders, p_bottom_pos=(0,0), p_init_lat_spd=0, p_init_lng_spd=0, p_max_spd=30, is_motorcycle=False, set_top=False):
        self.lane_borders = lane_borders
        # global target_vy
        # configure image for sprite
        pygame.sprite.Sprite. __init__(self)  # must do this
        self.image_path = p_image
        self.image = pygame.image.load(p_image).convert_alpha()
        h_width, h_height = self.image.get_size()

        # change the size of object to fit on the road lanes based on its type
        self.is_motorcycle = is_motorcycle
        if is_motorcycle:
            self.image = pygame.transform.scale(self.image, (globals.ROAD_LANE_WIDTH * 0.3, globals.ROAD_LANE_WIDTH * (h_height / h_width) * 0.3))  # scale the image based on given globals.WIDTH
        else:
            self.image = pygame.transform.scale(self.image, (globals.ROAD_LANE_WIDTH * 0.6, globals.ROAD_LANE_WIDTH * (h_height / h_width) * 0.6))  # scale the image based on given globals.WIDTH

        # save current rect position of the object
        self.rect = self.image.get_rect()  # rectangle that enclose the sprite..defines how wide and tall

        # initialize starting position
        self.rect.centerx = p_bottom_pos[0]
        self.rect.bottom = p_bottom_pos[1]
        if set_top:
            self.rect.top = p_bottom_pos[1]

        # initialize starting speed to zero
        self.vx = p_init_lat_spd
        self.vy = p_init_lng_spd
        self.max_vy = p_max_spd

        self.steering_wheel_angle = 0
        self.turning_speed = 50  # Increase this value to make the circular motion smoother
        self.turning_radius = 0
        self.initial_x = self.rect.centerx  # Store the initial position in case of straight movement
        self.initial_y = self.rect.centery
        self.angle_direction = 1
        self.angle = 0
        self.direction = 1
        self.initial_angle = 0
        self.initial_radius = 0

        self.rot_img = []
        self.min_angle = 1
        for i in range(360):
            # This rotation has to match the angle in radians later
            # So offet the angle (0 degrees = "north") by 90° to be angled 0-radians (so 0 rad is "east")
            rotated_image = pygame.transform.rotozoom(self.image, (i * self.min_angle), 1)
            self.rot_img.append(rotated_image)

    # --------------------------- GET INFO METHODS --------------------------

    def get_speed(self):
        return self.vy

    def get_rect_position(self):
        return self.rect

    def get_bottom_position(self):
        return self.rect.centerx, self.rect.bottom

    def get_lane(self):
        return get_lane(self.rect.centerx, self.lane_borders)

    def get_image_path(self):
        return self.image_path

    def get_size(self):
        return self.image.get_size()

    def get_steering_wheel_angle(self):
        return self.steering_wheel_angle

    # ---------------- SET/ACTION METHODS --------------------
    def set_speed(self, p_spd):
        self.vy = int(p_spd)

    def go_to_lane(self, p_lane_num):
        self.rect.centerx = get_lane_coordinate(p_lane_num, self.lane_borders)

    def speed_up(self, p_spd_inc=1):
        if self.vy < self.max_vy:
            self.vy += int(p_spd_inc)

    def speed_down(self, p_spd_inc=1):
        self.vy -= int(p_spd_inc)

    def stop(self):
        self.vy = 0

    def move_left(self):
        self.go_to_lane(self.get_lane() - 1)

    def move_right(self):
        self.go_to_lane(self.get_lane() + 1)

    # parallel parking methods

    def turn_right(self, degree_step=1):
        self.initial_angle = self.angle
        self.initial_radius = self.turning_radius
        self.steering_wheel_angle = max(self.steering_wheel_angle - degree_step, -globals.MAX_STEERING_ANGLE)

    def turn_left(self, degree_step=1):
        self.initial_angle = self.angle
        self.initial_radius = self.turning_radius
        self.steering_wheel_angle = min(self.steering_wheel_angle + degree_step, +globals.MAX_STEERING_ANGLE)

    def forward(self, speed=5):
        self.direction = 1

        if self.steering_wheel_angle == 0:
            # Update initial position when moving straight
            self.initial_x = self.rect.centerx
            self.initial_y = self.rect.centery

            # Move the car in a straight line without turning
            self.rect.centerx += self.direction * speed * math.sin(math.radians(self.angle))
            self.rect.centery -= self.direction * speed * math.cos(math.radians(self.angle))
        else:
            # Calculate turning radius based on steering angle and forward speed
            self.turning_radius = globals.TURNING_RADIUS_MULTIPLIER / math.radians(self.steering_wheel_angle)

            # Update rotation angle
            rotation_angle = self.turning_speed / abs(self.turning_radius)
            if globals.DEBUG:
                print(f"direct - initial angle: {math.degrees(self.initial_angle)} - angle: {math.degrees(self.angle)} - rotation angle: {rotation_angle} -  radius: {self.turning_radius} - steering: {self.steering_wheel_angle}")

            if self.initial_angle == 0:
                if self.steering_wheel_angle > 0:
                    self.angle -= rotation_angle
                    self.rect.centerx = self.initial_x + self.turning_radius * math.cos(math.radians(self.angle)) - self.turning_radius
                    self.rect.centery = self.initial_y + self.turning_radius * math.sin(math.radians(self.angle))
                else:
                    self.angle -= rotation_angle
                    self.rect.centerx = self.initial_x + self.turning_radius * math.cos(math.radians(self.angle)) - self.turning_radius
                    self.rect.centery = self.initial_y - self.turning_radius * math.sin(math.radians(self.angle))
            else:
                self.angle += rotation_angle
                self.rect.centerx = self.initial_x + self.turning_radius * math.cos(math.radians(self.angle))  # - self.turning_radius * math.cos(math.radians(self.initial_angle))
                self.rect.centery = self.initial_y + self.turning_radius * math.sin(math.radians(self.angle))  # - self.turning_radius * math.sin(math.radians(self.initial_angle))

    def reverse(self, speed=5):
        self.direction = -1
        if self.steering_wheel_angle == 0:
            # Update initial position when moving straight
            self.initial_x = self.rect.centerx
            self.initial_y = self.rect.centery

            # Move the car in a straight line without turning

            if self.angle != 0:
                self.rect.centerx += self.direction * speed * math.sin(math.radians(-self.angle))
                self.rect.centery -= self.direction * speed * math.cos(math.radians(-self.angle))
            else:
                self.rect.centerx += self.direction * speed * math.sin(math.radians(self.angle))
                self.rect.centery -= self.direction * speed * math.cos(math.radians(self.angle))

        else:
            # Calculate turning radius based on steering angle and forward speed
            self.turning_radius = globals.TURNING_RADIUS_MULTIPLIER / math.radians(self.steering_wheel_angle)

            # Update rotation angle
            rotation_angle = self.turning_speed / abs(self.turning_radius)
            if globals.DEBUG:
                print(f"reverse - angle: {math.degrees(self.angle)} - rotation angle: {rotation_angle} -  radius: {self.turning_radius} - steering: {self.steering_wheel_angle}")
            if self.initial_angle == 0:
                if self.steering_wheel_angle > 0:
                    self.angle += rotation_angle
                    self.rect.centerx = self.initial_x + self.turning_radius * math.cos(math.radians(self.angle)) - self.turning_radius + self.initial_radius
                    self.rect.centery = self.initial_y + self.turning_radius * math.sin(math.radians(self.angle))
                else:
                    self.angle += rotation_angle
                    self.rect.centerx = self.initial_x + self.turning_radius * math.cos(math.radians(self.angle)) - self.turning_radius + self.initial_radius
                    self.rect.centery = self.initial_y - self.turning_radius * math.sin(math.radians(self.angle))
            else:
                if self.steering_wheel_angle == 0:
                    self.angle += rotation_angle
                    self.rect.centerx = self.initial_x + self.turning_radius * math.cos(math.radians(self.angle)) - self.turning_radius + self.initial_radius
                    self.rect.centery = self.initial_y + self.turning_radius * math.sin(math.radians(self.angle))
                else:
                    self.angle -= rotation_angle
                    self.rect.centerx = self.initial_x + self.turning_radius * math.cos(math.radians(self.angle)) - self.turning_radius * math.cos(math.radians(self.initial_angle))
                    self.rect.centery = self.initial_y - self.turning_radius * math.sin(math.radians(self.angle)) + self.turning_radius * math.sin(math.radians(self.initial_angle))

    # renders the host vehicle on the display

    def parrallel_parking_render(self, screen):
        if self.initial_angle == 0:
            if self.direction == 1:
                if self.steering_wheel_angle > 0:
                    self.angle_direction = -1
                else:
                    self.angle_direction = 1
            else:
                if self.steering_wheel_angle > 0:
                    self.angle_direction = -1
                else:
                    self.angle_direction = 1

        rotated_car = pygame.transform.rotate(self.image, self.angle_direction * self.angle)

        car_rect = rotated_car.get_rect(center=(self.rect.centerx, self.rect.centery))
        screen.blit(rotated_car, car_rect.topleft)

    # ---------------------- SENSOR METHODS -----------------------

    def is_visible(self):
        return is_on_screen(self.rect.centerx, self.rect.top)

    def is_left_lane_available(self):
        return is_lane_available(self.rect.centerx, self.lane_borders, "left")

    def is_right_lane_available(self):
        return is_lane_available(self.rect.centerx, self.lane_borders, "right")

    def sensing(self, targets):
        inf = float('inf')
        perception = edict({
            "forward_target":{
                "vy": inf,
                "dist": inf
            },
            "left_forward_target":{
                "vy": inf,
                "dist": inf
            },
            "right_forward_target":{
                "vy": inf,
                "dist": inf
            },
            "collision": False,
            "running":True,
            "host_speed":self.get_speed(),
            "host_lane":self.get_lane(),
            "left_lane_available":self.is_left_lane_available(),
            "right_lane_available":self.is_right_lane_available()
        })

        for target in targets:
            # determine collision
            if self.rect.colliderect(target):
                perception["running"] = False
                perception.collision = True

            # foward perception
            if target.is_visible() and target.get_lane() == self.get_lane():
                # check if closest forward vehicle
                dist = self.rect.top - target.rect.bottom
                if dist < perception.forward_target.dist:
                    perception.forward_target.dist = dist
                    perception.forward_target.vy = target.vy

            # left perception
            if self.is_left_lane_available():
                if target.is_visible() and target.get_lane() == self.get_lane() - 1:
                    # check if closest left vehicle
                    dist = self.rect.top - target.rect.bottom
                    if dist < perception.left_forward_target.dist:
                        perception.left_forward_target.dist = dist
                        perception.left_forward_target.vy = target.vy

            # right perception
            if self.is_right_lane_available():
                if target.is_visible() and target.get_lane() == self.get_lane() + 1:
                    # check if closest right vehicle
                    dist = self.rect.top - target.rect.bottom
                    if dist < perception.right_forward_target.dist:
                        perception.right_forward_target.dist = dist
                        perception.right_forward_target.vy = target.vy

        return perception



from simulator.traffic import SingleTargetAhead
import random
import pygame
from simulator import globals


class TrafficRouter:
    def __init__(self, lane_borders, road_borders, num_lanes, gap_offset=0, speed_range=[3, 8],
                 traffic_object_start_point=0, has_speed_change=True, speed_change_counter_range=[200, 700], new_traffic_object_probability=0.3):
        # initialize the router parameters
        self.lane_borders = lane_borders
        self.road_borders = road_borders
        self.num_lanes = num_lanes
        self.speed_range = speed_range
        self.gap_offset = gap_offset
        self.traffic_object_start_point = traffic_object_start_point
        self.has_speed_change = has_speed_change
        self.speed_change_counter_range = speed_change_counter_range
        # initialize distance for changing speed randomly
        self.speed_change_selected_counter = random.randint(self.speed_change_counter_range[0], self.speed_change_counter_range[1])
        # create random speed for each lane
        self.lanes_speed = [random.randint(self.speed_range[0], self.speed_range[1]) for i in range(0, self.num_lanes)]

        # setup traffic group
        self.targets = pygame.sprite.Group()
        self.distance_pointer_list = [0] * self.num_lanes
        self.gap_size_list = [0] * self.num_lanes
        self.new_traffic_object_probability = new_traffic_object_probability

    def generate_random_traffic_object(self, lane_number):
        # select an image randomly
        image_idx = random.randint(0,4)

        # check if selected image is motorcycle
        is_motorcycle = image_idx > 3

        # create new traffic object
        new_target = SingleTargetAhead(globals.VEHICLE_IMAGES[image_idx], self.lane_borders,
                                       p_bottom_pos=(0,self.traffic_object_start_point),
                                       p_init_lng_spd=self.lanes_speed[lane_number], is_motorcycle=is_motorcycle, lane=lane_number)

        # get the globals.HEIGHT of the vehicle to find the gap_size for adding another traffic object
        new_target_width, new_target_height = new_target.get_size()
        self.gap_size_list[lane_number] = new_target_height + self.gap_offset + 10  # +10 to make sure there is at least a small gap between two consecutive objects
        self.targets.add(new_target)

    def change_lanes_speed(self, host_speed):
        # create new random speed array
        self.lanes_speed = [random.randint(self.speed_range[0], self.speed_range[1]) for i in range(0, self.num_lanes)]
        if globals.DEBUG:
            print(f"new speed: {self.lanes_speed}")

        # change the speed of previous traffic objects and update them
        for target in self.targets:
            # get previous item information
            lane_idx = target.get_lane()
            target.set_speed(self.lanes_speed[lane_idx])
            target.update(host_speed)

    def update(self, host_speed):
        # for the distance pointer we should consider the fastest traffic object (as it would be in the screen for longer time)
        for i in range(0, self.num_lanes):
            self.distance_pointer_list[i] += host_speed - self.lanes_speed[i]

            # if the specified gap size is travelled, add new traffic object
            if self.distance_pointer_list[i] >= self.gap_size_list[i]:
                self.distance_pointer_list[i] = 0

                # add new traffic object with random probability
                temp_random = random.random()
                if temp_random < self.new_traffic_object_probability:
                    self.generate_random_traffic_object(i)

        # if the specified distance size is travelled, change lanes speed
        if self.speed_change_selected_counter <= 0 and self.has_speed_change:
            # select a random distance for next speed change period
            self.speed_change_selected_counter = random.randint(self.speed_change_counter_range[0], self.speed_change_counter_range[1])
            if globals.DEBUG:
                print(f"new selected speed counter: {self.speed_change_selected_counter}")
            self.change_lanes_speed(host_speed)
        else:
            self.speed_change_selected_counter -= 1
            # update would occur inside change_lanes_speed whenever it is being called as it is more efficient
            for target in self.targets:
                target.update(host_speed)

        return self.targets, self.lanes_speed

    def render(self, screen):
        self.targets.draw(screen)



from simulator.vehicle import Vehicle
from simulator import globals


class SingleTargetAhead(Vehicle):
    def __init__(self, p_image, lane_borders, lane=0, p_bottom_pos=(0,0), p_init_lat_spd=0, p_init_lng_spd=0, p_max_spd=30, p_cycle_limit=1000, is_motorcycle=False, set_top=False):
        super().__init__(p_image, lane_borders, p_bottom_pos, p_init_lat_spd, p_init_lng_spd, p_max_spd, is_motorcycle, set_top=set_top)

        # set target speed
        self.set_speed(self.vy)

        # set target lane
        self.go_to_lane(lane)

        # simulation attributes
        self.status = 1  # (1; active, -1: terminate)
        self.cycle = 0
        self.cycle_limit = p_cycle_limit

    def update(self, p_host_spd):
        if self.status == 1:
            # update longitudinal position based on relative speed
            self.rect.centery = self.rect.centery + (p_host_spd - self.vy)
            # check if deactivate vehicle
            if self.rect.top > globals.HEIGHT or self.cycle > self.cycle_limit:
                self.status = -1

        if self.status == -1:
            self.terminate(p_host_spd)

    def terminate(self, p_host_spd):
        if self.rect.top > globals.HEIGHT:
            self.kill()
        else:
            self.vy = int(p_host_spd * 2)
            self.rect.centery = self.rect.centery + (p_host_spd - self.vy)
            if self.rect.bottom < 0:
                if globals.DEBUG:
                    print("kill")
                self.kill()


# I don't think that we need this class anymore
class MultipleTargetAhead:
    def __init__(self, p_images, lane_borders, number_of_vehicles=2, lane=0, p_bottom_pos=(0,0), p_init_lng_spd=0, p_max_spd=30, p_cycle_limit=1000, gap=globals.ROAD_LANE_WIDTH * 0.4):
        self.p_images = p_images
        self.lane_borders = lane_borders
        self.vehicle_list = []

        start_point = p_bottom_pos[1]
        other_vehicle_height = 0
        for i in range(number_of_vehicles):
            temp = SingleTargetAhead(self.p_images[i], self.lane_borders,
                                     p_bottom_pos=(0,start_point + other_vehicle_height),
                                     p_init_lng_spd=0, is_motorcycle=False, p_max_spd=p_max_spd, p_cycle_limit=p_cycle_limit)
            temp.set_speed(p_init_lng_spd)
            temp.go_to_lane(lane)

            temp_width, temp_height = temp.get_size()
            other_vehicle_height += gap + temp_height

            self.vehicle_list.append(temp)



import math
import random

import pygame

from simulator import globals
from simulator.utils import calculate_roadside_area_width, get_sidewalk_x_loc

# create horizontal globals.WHITE lines as well
# have a fixed globals.WIDTH road
# then add extra objects


class Env:
    def __init__(self, number_of_lanes, show_horizontal_line=False):
        # calculate and setup number of lanes, grass area globals.WIDTH, etc.
        self.number_of_lanes = number_of_lanes
        self.number_of_white_lines = self.number_of_lanes - 1
        self.roadside_area_widths = calculate_roadside_area_width(
            globals.WIDTH
            - (
                globals.ROAD_LANE_WIDTH * self.number_of_lanes
                + self.number_of_white_lines * globals.WHITE_LANE_WIDTH
            ),
            globals.GRASS_AREA_RATIO,
            globals.SIDEWALK_AREA_RATIO,
        )
        self.show_horizontal_line = show_horizontal_line

        # calculate the start and end points of scene rectangles (sidewalk, road)
        self.sidewalk_border_left = self.roadside_area_widths.grass_area_width
        self.sidewalk_border_right = (
            globals.WIDTH - self.roadside_area_widths.grass_area_width * 2
        )
        self.road_border_left = (
            self.roadside_area_widths.grass_area_width
            + self.roadside_area_widths.sidewalk_area_width
        )
        self.road_border_right = globals.WIDTH - 2 * (
            self.roadside_area_widths.grass_area_width
            + self.roadside_area_widths.sidewalk_area_width
        )

        # calculate the positions of globals.WHITE lines
        self.distance_pointer_white_lines = 0
        if show_horizontal_line:
            self.horizontal_line_x_loc = (
                globals.WIDTH
                - (
                    self.roadside_area_widths.grass_area_width
                    + self.roadside_area_widths.sidewalk_area_width
                )
                - globals.HORIZONTAL_LANE_WIDTH
            )
            self.horizontal_line_y_loc = range(
                globals.HORIZONTAL_LANE_START_Y,
                globals.HEIGHT,
                globals.HORIZONTAL_LANE_GAP + globals.HORIZONTAL_LANE_HEIGHT,
            )
            if globals.DEBUG:
                print(self.horizontal_line_y_loc)
        else:
            self.white_y_loc = range(
                -globals.WHITE_LANE_HEIGHT,
                globals.HEIGHT,
                math.floor(globals.WHITE_LANE_HEIGHT + globals.WHITE_LANE_GAP),
            )
            self.white_x_loc = range(
                math.floor(
                    self.roadside_area_widths.grass_area_width
                    + self.roadside_area_widths.sidewalk_area_width
                    + globals.ROAD_LANE_WIDTH
                ),
                math.floor(
                    globals.WIDTH
                    - (
                        self.roadside_area_widths.grass_area_width
                        + self.roadside_area_widths.sidewalk_area_width
                    )
                ),
                globals.ROAD_LANE_WIDTH + globals.WHITE_LANE_WIDTH,
            )

        # calculate the positions of trees
        self.distance_pointer_trees = 0
        self.trees_group = pygame.sprite.Group()
        for tree_idx in range(
            globals.TREES_START_POINT,
            globals.HEIGHT,
            math.floor(globals.TREES_WIDTH + globals.TREES_GAP),
        ):
            tree_right = Tree(
                globals.TREES_IMAGES[random.randint(0, 3)],
                globals.TREES_WIDTH,
                (self.roadside_area_widths.grass_area_width, tree_idx),
            )
            tree_left = Tree(
                globals.TREES_IMAGES[random.randint(0, 3)],
                globals.TREES_WIDTH,
                (globals.WIDTH - self.roadside_area_widths.grass_area_width, tree_idx),
            )
            self.trees_group.add(tree_right)
            self.trees_group.add(tree_left)

        # calculate the positions of houses
        self.distance_pointer_houses = 0
        self.houses_group = pygame.sprite.Group()
        for house_idx in range(
            globals.HOUSE_START_POINT,
            globals.HEIGHT,
            math.floor(globals.HOUSES_WIDTH + globals.HOUSES_GAP),
        ):
            house_right = House(
                globals.HOUSES_IMAGES[random.randint(0, 3)],
                globals.HOUSES_WIDTH,
                (0, house_idx),
            )
            house_left = Tree(
                globals.HOUSES_IMAGES[random.randint(0, 3)],
                globals.HOUSES_WIDTH,
                (globals.WIDTH, house_idx),
            )
            self.houses_group.add(house_right)
            self.houses_group.add(house_left)

        # calculate the positions of sidewalk objects
        self.distance_pointer_sidewalk = 0
        self.sidewalk_group = pygame.sprite.Group()
        for sidewalk_idx in range(
            globals.SIDEWALK_OBJECTS_START_POINT,
            globals.HEIGHT,
            math.floor(
                globals.SIDEWALK_OBJECTS_WIDTH
                + random.randint(
                    globals.SIDEWALK_OBJECTS_GAP_MIN, globals.SIDEWALK_OBJECTS_GAP_MAX
                )
            ),
        ):
            right_idx = random.randint(0, 7)
            left_idx = random.randint(0, 7)

            # is the object in right lane or left lane of the sidewalk?
            right_pos = random.randint(0, 1)
            left_pos = random.randint(0, 1)

            swo_right = SidewalkObject(
                globals.SIDEWALK_OBJECTS_IMAGES[right_idx],
                globals.SIDEWALK_OBJECTS_WIDTH,
                0 if right_idx < 4 else 1,
                (
                    get_sidewalk_x_loc(
                        globals.WIDTH,
                        self.roadside_area_widths.grass_area_width,
                        globals.GRASS_AREA_RATIO,
                        globals.SIDEWALK_AREA_RATIO,
                        right_pos,
                    ),
                    sidewalk_idx,
                ),
                sidewalk_loc=right_pos,
            )
            swo_left = SidewalkObject(
                globals.SIDEWALK_OBJECTS_IMAGES[left_idx],
                globals.SIDEWALK_OBJECTS_WIDTH,
                0 if right_idx < 4 else 1,
                (
                    get_sidewalk_x_loc(
                        globals.WIDTH,
                        self.roadside_area_widths.grass_area_width,
                        globals.GRASS_AREA_RATIO,
                        globals.SIDEWALK_AREA_RATIO,
                        left_pos,
                        left_side=True,
                    ),
                    sidewalk_idx,
                ),
                sidewalk_loc=left_pos,
            )
            self.sidewalk_group.add(swo_right)
            self.sidewalk_group.add(swo_left)

    def get_horizontal_line_location(self):
        temp = []
        for y_loc in self.horizontal_line_y_loc:
            temp.append((self.horizontal_line_x_loc, y_loc))
        return temp

    def get_vertical_line_location(self):
        return list(self.white_x_loc)

    def get_sidwalk_road_borders(self):
        return [
            self.roadside_area_widths.grass_area_width
            + self.roadside_area_widths.sidewalk_area_width,
            globals.WIDTH
            - (
                self.roadside_area_widths.grass_area_width
                + self.roadside_area_widths.sidewalk_area_width
            ),
        ]

    def update(self, speed):
        if globals.DEBUG:
            print(f"speed inside env: {speed}")
        # update globals.WHITE lines
        if not self.show_horizontal_line:
            temp_y_list = []
            if self.distance_pointer_white_lines > (
                globals.WHITE_LANE_GAP + globals.WHITE_LANE_HEIGHT
            ):
                self.distance_pointer_white_lines = 0
                temp_y_list.append(-globals.WHITE_LANE_HEIGHT)

            for y_loc in self.white_y_loc:
                temp = y_loc + speed
                if temp <= globals.HEIGHT:
                    temp_y_list.append(temp)

            self.white_y_loc = temp_y_list
        else:
            speed = 0

        # update trees
        if self.distance_pointer_trees > (globals.TREES_GAP + globals.TREES_WIDTH):
            self.distance_pointer_trees = 0
            tree_right = Tree(
                globals.TREES_IMAGES[random.randint(0, 3)],
                globals.TREES_WIDTH,
                (self.roadside_area_widths.grass_area_width, globals.TREES_START_POINT),
            )
            tree_left = Tree(
                globals.TREES_IMAGES[random.randint(0, 3)],
                globals.TREES_WIDTH,
                (
                    globals.WIDTH - self.roadside_area_widths.grass_area_width,
                    globals.TREES_START_POINT,
                ),
            )
            self.trees_group.add(tree_right)
            self.trees_group.add(tree_left)

        for tree in self.trees_group:
            tree.update(speed)

        # update houses
        if self.distance_pointer_houses > (globals.HOUSES_WIDTH + globals.HOUSES_GAP):
            self.distance_pointer_houses = 0
            house_right = House(
                globals.HOUSES_IMAGES[random.randint(0, 3)],
                globals.HOUSES_WIDTH,
                (0, globals.HOUSE_START_POINT),
            )
            house_left = House(
                globals.HOUSES_IMAGES[random.randint(0, 3)],
                globals.HOUSES_WIDTH,
                (globals.WIDTH, globals.HOUSE_START_POINT),
            )
            self.houses_group.add(house_right)
            self.houses_group.add(house_left)

        for house in self.houses_group:
            house.update(speed)

        # update sidewalk objects
        if self.distance_pointer_sidewalk > (
            globals.SIDEWALK_OBJECTS_WIDTH
            + random.randint(
                globals.SIDEWALK_OBJECTS_GAP_MIN, globals.SIDEWALK_OBJECTS_GAP_MAX
            )
        ):
            self.distance_pointer_sidewalk = 0
            right_idx = random.randint(0, 7)
            left_idx = random.randint(0, 7)

            # is the object in right lane or left lane of the sidewalk?
            right_pos = random.randint(0, 1)
            left_pos = random.randint(0, 1)

            swo_right = SidewalkObject(
                globals.SIDEWALK_OBJECTS_IMAGES[right_idx],
                globals.SIDEWALK_OBJECTS_WIDTH,
                0 if right_idx < 4 else 1,
                (
                    get_sidewalk_x_loc(
                        globals.WIDTH,
                        self.roadside_area_widths.grass_area_width,
                        globals.GRASS_AREA_RATIO,
                        globals.SIDEWALK_AREA_RATIO,
                        right_pos,
                    ),
                    globals.SIDEWALK_OBJECTS_START_POINT,
                ),
                sidewalk_loc=right_pos,
            )
            swo_left = SidewalkObject(
                globals.SIDEWALK_OBJECTS_IMAGES[left_idx],
                globals.SIDEWALK_OBJECTS_WIDTH,
                0 if right_idx < 4 else 1,
                (
                    get_sidewalk_x_loc(
                        globals.WIDTH,
                        self.roadside_area_widths.grass_area_width,
                        globals.GRASS_AREA_RATIO,
                        globals.SIDEWALK_AREA_RATIO,
                        left_pos,
                        left_side=True,
                    ),
                    globals.SIDEWALK_OBJECTS_START_POINT,
                ),
                sidewalk_loc=left_pos,
            )
            self.sidewalk_group.add(swo_right)
            self.sidewalk_group.add(swo_left)

        for swo in self.sidewalk_group:
            swo.update(speed)

        # update itr_number for each object group
        self.distance_pointer_white_lines += speed
        self.distance_pointer_trees += speed
        self.distance_pointer_houses += speed
        self.distance_pointer_sidewalk += speed

    def render(self, window, show_sidewalk_objects=True):
        # re-render the road, grass and sidewalk area
        pygame.draw.rect(
            window, globals.GRASS_COLOR, (0, 0, globals.WIDTH, globals.HEIGHT)
        )
        pygame.draw.rect(
            window,
            globals.SIDEWALK_COLOR,
            (self.sidewalk_border_left, 0, self.sidewalk_border_right, globals.HEIGHT),
        )
        pygame.draw.rect(
            window,
            globals.ASPHALT_COLOR,
            (self.road_border_left, 0, self.road_border_right, globals.HEIGHT),
        )

        # draw globals.WHITE/globals.YELLOW lines
        if not self.show_horizontal_line:
            for x_loc in self.white_x_loc:
                for y_loc in self.white_y_loc:
                    if y_loc < globals.HEIGHT:
                        pygame.draw.rect(
                            window,
                            globals.WHITE,
                            (
                                x_loc,
                                y_loc,
                                globals.WHITE_LANE_WIDTH,
                                globals.WHITE_LANE_HEIGHT,
                            ),
                        )
        else:
            for y_loc in self.horizontal_line_y_loc:
                pygame.draw.rect(
                    window,
                    globals.YELLOW,
                    (
                        self.horizontal_line_x_loc,
                        y_loc,
                        globals.HORIZONTAL_LANE_WIDTH,
                        globals.HORIZONTAL_LANE_HEIGHT,
                    ),
                )

        # draw sidewalk objects
        if show_sidewalk_objects:
            self.sidewalk_group.draw(window)

        # draw trees
        self.trees_group.draw(window)

        # draw houses
        self.houses_group.draw(window)

    def reset(self):
        self.distance_pointer_white_lines = 0


class Tree(pygame.sprite.Sprite):
    def __init__(self, p_image, WIDTH, p_bottom_pos=(0, 0)):
        pygame.sprite.Sprite.__init__(self)  # must do this
        self.image = pygame.transform.scale(
            pygame.image.load(p_image).convert_alpha(), (WIDTH, WIDTH)
        )
        self.rect = (
            self.image.get_rect()
        )  # rectangle that enclose the sprite..defines how wide and tall
        self.rect.centerx = p_bottom_pos[0]
        self.rect.bottom = p_bottom_pos[1]

    def update(self, p_spd):
        self.vy = p_spd
        self.rect.y += self.vy
        if self.rect.top >= globals.HEIGHT:
            self.kill()


class House(pygame.sprite.Sprite):
    def __init__(self, p_image, WIDTH, p_bottom_pos=(0, 0)):
        pygame.sprite.Sprite.__init__(self)  # must do this
        self.image = pygame.transform.scale(
            pygame.image.load(p_image).convert_alpha(), (WIDTH, WIDTH)
        )
        self.rect = (
            self.image.get_rect()
        )  # rectangle that enclose the sprite..defines how wide and tall
        self.rect.centerx = p_bottom_pos[0]
        self.rect.bottom = p_bottom_pos[1]

    def update(self, p_spd):
        self.vy = p_spd
        self.rect.y += self.vy
        if self.rect.top >= globals.HEIGHT:
            self.kill()


class SidewalkObject(pygame.sprite.Sprite):
    def __init__(self, p_image, WIDTH, obj_type, p_bottom_pos=(0, 0), sidewalk_loc=0):
        pygame.sprite.Sprite.__init__(self)  # must do this
        self.image = pygame.image.load(p_image).convert_alpha()
        h_width, h_height = self.image.get_size()
        self.object_type = 0
        if obj_type == 0:  # pedestrian
            self.image = pygame.transform.scale(
                pygame.image.load(p_image).convert_alpha(),
                (WIDTH * 1.2, WIDTH * (h_height / h_width) * 1.3),
            )  # scale the image based on given globals.WIDTH
        else:  # bike
            self.object_type = 1
            self.image = pygame.transform.scale(
                pygame.image.load(p_image).convert_alpha(),
                (WIDTH, WIDTH * (h_height / h_width)),
            )  # scale the image based on given globals.WIDTH

        self.sidewalk_loc = 0
        if sidewalk_loc == 1:  # check that object is in which side of the sidewalk
            self.sidewalk_loc = 1
            self.image = pygame.transform.rotate(
                self.image, 180
            )  # because they should be in the opposite direction

        self.rect = (
            self.image.get_rect()
        )  # rectangle that enclose the sprite..defines how wide and tall
        self.rect.centerx = p_bottom_pos[0]
        self.rect.bottom = p_bottom_pos[1]

    def update(self, p_spd):
        ratio = 1
        if self.sidewalk_loc == 1:
            ratio = -1
        if self.object_type == 0:
            self.vy = p_spd + globals.PEDESTRIAN_SPEED * ratio
        else:
            self.vy = p_spd + globals.BIKE_SPEED * ratio
        self.rect.y += self.vy
        if self.rect.top >= globals.HEIGHT:
            self.kill()



from setuptools import find_packages, setup

setup(
    name='veh_sim',
    version='0.3.0',
    description='A short description of your package',
    author='Richard Hamilton',
    author_email='r.hamilton@mathstronauts.ca',
    url='https://github.com/your_username/your_package_name',
    include_package_data=True,
    package_data={'': ['images/*']},
    packages=find_packages(),
    install_requires=[
        "pygame",
        "easydict"
    ],
    extras_require={
        "dev": [
            "black>=23.1a1",
            "flake8>=3.8.4",
            "pytest>=6.2.5",
            "pytest-cov>=2.11.1",
            "isort>=5.7.0",
            "pre-commit>=2.9.3",
            "ipython>=8.8.0",
            "ipykernel>=6.20.2",
            "ipywidgets>=7.6.3",
            "build",
            "twine",
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



from easydict import EasyDict as edict
import json
import os

# read configuration file
cfg_file = os.path.join(os.path.dirname(__file__),"simulation_cfg.json")
with open(cfg_file, "r") as f:
    sim_cfg = edict(json.load(f))

FPS = sim_cfg.screen.FPS


# define colors
ASPHALT_COLOR = (42, 41, 34)
SIDEWALK_COLOR = (183, 173, 63)
GRASS_COLOR = (0,154,23)
WHITE = (255,255,255)
BLACK = (0,0,0)
RED = (255,0,0)
YELLOW = (255, 255, 0)
GREEN = (0,255,0)
BLUE = (0,0,255)
GRAY = (128, 128, 128)

# window sizes
# you can change the screen size here
WIDTH = 1100
HEIGHT = 600
FONT_SIZE = 32
PANEL_FONT_SIZE = 10
# PARALLEL_PARKING_STEPS = [63, 98, 190, 233, 255, 290, 385]#small
PARALLEL_PARKING_STEPS = [99, 134, 254, 289, 334, 369, 489]  # medium

# base size (don't change this size)
# base size => WIDTH: 1100, HEIGHT: 600
BASE_WIDTH = 1100
BASE_HEIGHT = 600

# road
WHITE_LANE_WIDTH = round(sim_cfg.road.whiteLaneWidth * WIDTH / BASE_WIDTH)
WHITE_LANE_HEIGHT = round(sim_cfg.road.whiteLaneHeight * HEIGHT / BASE_HEIGHT)
WHITE_LANE_GAP = round(sim_cfg.road.whiteLaneGap * HEIGHT / BASE_HEIGHT)
ROAD_LANE_WIDTH = round(sim_cfg.road.laneWidth * WIDTH / BASE_WIDTH)
HORIZONTAL_LANE_WIDTH = round(sim_cfg.road.horizontalLaneWidth * WIDTH / BASE_WIDTH)
HORIZONTAL_LANE_HEIGHT = round(sim_cfg.road.horizontalLaneThickness * HEIGHT / BASE_HEIGHT)
HORIZONTAL_LANE_GAP = round(sim_cfg.road.horizontalLaneGap * HEIGHT / BASE_HEIGHT)
HORIZONTAL_LANE_START_Y = round(sim_cfg.road.horizontalLaneStartY * HEIGHT / BASE_HEIGHT)
# LANE_BORDERS = [] I should use global variable for the lane borders in the Vehicle class later

# road sides
GRASS_AREA_RATIO = sim_cfg.road.GRASS_AREA_RATIO  # ratio of grass area WIDTH in the calculation
SIDEWALK_AREA_RATIO = sim_cfg.road.SIDEWALK_AREA_RATIO  # ratio of sidewalk area WIDTH in the calculation
# for example: grass_area_width = (window - n_lanes * lane_width)/(2 * (GRASS_AREA_RATIO + SIDEWALK_AREA_RATIO))

# trees
TREES_WIDTH = round(sim_cfg.env_objects.TREES_WIDTH * WIDTH / BASE_WIDTH)
TREES_GAP = round(sim_cfg.env_objects.TREES_GAP * HEIGHT / BASE_HEIGHT)
TREES_START_POINT = round(sim_cfg.env_objects.TREES_START_POINT * HEIGHT / BASE_HEIGHT)
TREES_IMAGES = [sim_cfg.env_objects.images.tree_0, sim_cfg.env_objects.images.tree_1, sim_cfg.env_objects.images.tree_2, sim_cfg.env_objects.images.tree_3]
# houses
HOUSES_WIDTH = round(sim_cfg.env_objects.house_width * WIDTH / BASE_WIDTH)
HOUSES_GAP = round(sim_cfg.env_objects.house_gap * HEIGHT / BASE_HEIGHT)
HOUSE_START_POINT = round(sim_cfg.env_objects.HOUSE_START_POINT * HEIGHT / BASE_HEIGHT)
HOUSES_IMAGES = [sim_cfg.env_objects.images.house_0, sim_cfg.env_objects.images.house_1, sim_cfg.env_objects.images.house_2, sim_cfg.env_objects.images.house_3]


# sidewalk objects
SIDEWALK_OBJECTS_WIDTH = round(sim_cfg.env_objects.sidewalk_object_width * WIDTH / BASE_WIDTH)
SIDEWALK_OBJECTS_GAP_MIN = round(sim_cfg.env_objects.sidewalk_object_gap_min * HEIGHT / BASE_HEIGHT)
SIDEWALK_OBJECTS_GAP_MAX = round(sim_cfg.env_objects.sidewalk_object_gap_max * HEIGHT / BASE_HEIGHT)
SIDEWALK_OBJECTS_START_POINT = round(sim_cfg.env_objects.sidewalk_object_start_point * HEIGHT / BASE_HEIGHT)
SIDEWALK_OBJECTS_IMAGES = [sim_cfg.env_objects.images.sidewalk_object_0, sim_cfg.env_objects.images.sidewalk_object_1, sim_cfg.env_objects.images.sidewalk_object_2, sim_cfg.env_objects.images.sidewalk_object_3, sim_cfg.env_objects.images.sidewalk_object_4, sim_cfg.env_objects.images.sidewalk_object_5, sim_cfg.env_objects.images.sidewalk_object_6, sim_cfg.env_objects.images.sidewalk_object_7]
BIKE_SPEED = sim_cfg.env_objects.BIKE_SPEED
PEDESTRIAN_SPEED = sim_cfg.env_objects.PEDESTRIAN_SPEED

# ------ CONSTANT TEXT MESSAGES ------ #
CAPTION = sim_cfg.screen.CAPTION


# VEHICLES
VEHICLES = edict(sim_cfg.vehicle)
VEHICLE_IMAGES = [sim_cfg.vehicle.images.car_0, sim_cfg.vehicle.images.car_1, sim_cfg.vehicle.images.car_2, sim_cfg.vehicle.images.car_3, sim_cfg.vehicle.images.motorcycle_0]

# host vehicle
HOST_VEHICLE = edict(sim_cfg.HOST_VEHICLE)
HOST_VEHICLE_IMAGES = sim_cfg.HOST_VEHICLE.image


# parameter panel
PADDING = 5

PARAMETER_PANEL_TEXT_HEIGHT = 20
PARAMETER_PANEL_BLACK_RECT_WIDTH = 150
PARAMETER_PANEL_YELLOW_RECT_WIDTH = PARAMETER_PANEL_BLACK_RECT_WIDTH + 2 * PADDING
PARAMETER_PANEL_BLACK_RECT_START_y = 30
PARAMETER_PANEL_YELLOW_RECT_START_y = 30 - PADDING
PARAMETER_PANEL_BLACK_RECT_HEIGHT = 15 * PARAMETER_PANEL_TEXT_HEIGHT  # 15 is number of lines (len(data)) in the pannel
PARAMETER_PANEL_YELLOW_RECT_HEIGHT = 15 * PARAMETER_PANEL_TEXT_HEIGHT + 2 * PADDING  # 15 is number of lines (len(data)) in the pannel

PARAMETER_PANEL_LEFT_MARGIN = 50
PARAMETER_PANEL_BLACK_RECT_START_X = WIDTH - PARAMETER_PANEL_BLACK_RECT_WIDTH - PARAMETER_PANEL_LEFT_MARGIN
PARAMETER_PANEL_TEXT_START_X = WIDTH - PARAMETER_PANEL_BLACK_RECT_WIDTH - PARAMETER_PANEL_LEFT_MARGIN + PADDING
PARAMETER_PANEL_YELLOW_RECT_START_X = WIDTH - PARAMETER_PANEL_BLACK_RECT_WIDTH - PARAMETER_PANEL_LEFT_MARGIN - PADDING


# PARALLEL PARKING
MAX_STEERING_ANGLE = 45


# ODOMETER parameters
ODOMETER_PANEL_BORDER_WIDTH = 5
ODOMETER_PANEL_START_POINT = 10
ODOMETER_PANEL_WIDTH = 200
ODOMETER_PANEL_HEIGHT = 75
ODOMETER_FONT_SIZE = 15

STEPS_IMAGES = sim_cfg.parallel_parking.step_images
CHECK_IMAGE = sim_cfg.parallel_parking.check_image
X_IMAGE = sim_cfg.parallel_parking.x_image


STEPS_IMAGES = sim_cfg.parallel_parking.step_images
print(len(STEPS_IMAGES))
CHECK_IMAGE = sim_cfg.parallel_parking.check_image
X_IMAGE = sim_cfg.parallel_parking.x_image


MAX_STEERING_ANGLE = 35
TURNING_RADIUS_MULTIPLIER = 100

PARKING_PANEL_START_POINT = 10
STEP_SIZE = (200, 50)
CHECK_SIZE = (40, 40)
STEP_GAP = 60


DEBUG = False


def set_screen_size(size="medium"):
    global HEIGHT, WIDTH
    global WHITE_LANE_WIDTH, WHITE_LANE_HEIGHT, WHITE_LANE_GAP, ROAD_LANE_WIDTH, HORIZONTAL_LANE_WIDTH
    global HORIZONTAL_LANE_HEIGHT, HORIZONTAL_LANE_GAP, HORIZONTAL_LANE_START_Y, TREES_WIDTH, TREES_GAP
    global REES_START_POINT, TREES_START_POINT, HOUSES_WIDTH, HOUSES_GAP, HOUSE_START_POINT, SIDEWALK_OBJECTS_WIDTH
    global SIDEWALK_OBJECTS_GAP_MIN, SIDEWALK_OBJECTS_GAP_MAX, SIDEWALK_OBJECTS_START_POINT, PARALLEL_PARKING_STEPS
    global TURNING_RADIUS_MULTIPLIER, STEP_SIZE, CHECK_SIZE, STEP_GAP, FONT_SIZE
    global PARAMETER_PANEL_BLACK_RECT_START_X, PARAMETER_PANEL_TEXT_START_X, PARAMETER_PANEL_YELLOW_RECT_START_X
    global PARAMETER_PANEL_YELLOW_RECT_WIDTH, PARAMETER_PANEL_LEFT_MARGIN, PARAMETER_PANEL_BLACK_RECT_START_y, PARAMETER_PANEL_YELLOW_RECT_START_y
    global PARAMETER_PANEL_BLACK_RECT_WIDTH, PANEL_FONT_SIZE, PARAMETER_PANEL_TEXT_HEIGHT, PARAMETER_PANEL_BLACK_RECT_HEIGHT
    global PARAMETER_PANEL_YELLOW_RECT_HEIGHT, ODOMETER_FONT_SIZE, ODOMETER_PANEL_WIDTH, ODOMETER_PANEL_START_POINT
    if size == "small":
        PARALLEL_PARKING_STEPS = [50, 85, 140, 175, 200, 235, 290]  # small
        WIDTH = 550
        HEIGHT = 300
        STEP_SIZE = (100, 30)
        CHECK_SIZE = (25, 25)
        STEP_GAP = 30
        FONT_SIZE = 24
        PANEL_FONT_SIZE = 8

        PARAMETER_PANEL_BLACK_RECT_WIDTH = 100
        PARAMETER_PANEL_LEFT_MARGIN = 10
        PARAMETER_PANEL_YELLOW_RECT_WIDTH = PARAMETER_PANEL_BLACK_RECT_WIDTH + 2 * PADDING
        PARAMETER_PANEL_TEXT_HEIGHT = 13
        PARAMETER_PANEL_BLACK_RECT_START_y = 5
        PARAMETER_PANEL_YELLOW_RECT_START_y = PARAMETER_PANEL_BLACK_RECT_START_y - PADDING
        PARAMETER_PANEL_BLACK_RECT_HEIGHT = 15 * PARAMETER_PANEL_TEXT_HEIGHT  # 15 is number of lines (len(data)) in the pannel
        PARAMETER_PANEL_YELLOW_RECT_HEIGHT = 15 * PARAMETER_PANEL_TEXT_HEIGHT + 2 * PADDING  # 15 is number of lines (len(data)) in the pannel
        ODOMETER_FONT_SIZE = 8
        ODOMETER_PANEL_WIDTH = 90
        ODOMETER_PANEL_START_POINT = 5
    elif size == "medium":
        PARALLEL_PARKING_STEPS = [99, 134, 254, 289, 334, 369, 489]  # medium
        WIDTH = 1100
        HEIGHT = 600
        FONT_SIZE = 32
    elif size == "large":
        FONT_SIZE = 36
        PARALLEL_PARKING_STEPS = [160, 195, 340, 375, 485, 520, 665]  # large
        WIDTH = 1650
        HEIGHT = 900
        STEP_SIZE = (250, 75)
        CHECK_SIZE = (60, 60)
        STEP_GAP = 75

    PARAMETER_PANEL_BLACK_RECT_START_X = WIDTH - PARAMETER_PANEL_BLACK_RECT_WIDTH - PARAMETER_PANEL_LEFT_MARGIN
    PARAMETER_PANEL_TEXT_START_X = WIDTH - PARAMETER_PANEL_BLACK_RECT_WIDTH - PARAMETER_PANEL_LEFT_MARGIN + PADDING
    PARAMETER_PANEL_YELLOW_RECT_START_X = WIDTH - PARAMETER_PANEL_BLACK_RECT_WIDTH - PARAMETER_PANEL_LEFT_MARGIN - PADDING

    WHITE_LANE_WIDTH = round(sim_cfg.road.whiteLaneWidth * WIDTH / BASE_WIDTH)
    WHITE_LANE_HEIGHT = round(sim_cfg.road.whiteLaneHeight * HEIGHT / BASE_HEIGHT)
    WHITE_LANE_GAP = round(sim_cfg.road.whiteLaneGap * HEIGHT / BASE_HEIGHT)
    ROAD_LANE_WIDTH = round(sim_cfg.road.laneWidth * WIDTH / BASE_WIDTH)
    HORIZONTAL_LANE_WIDTH = round(sim_cfg.road.horizontalLaneWidth * WIDTH / BASE_WIDTH)
    HORIZONTAL_LANE_HEIGHT = round(sim_cfg.road.horizontalLaneThickness * HEIGHT / BASE_HEIGHT)
    HORIZONTAL_LANE_GAP = round(sim_cfg.road.horizontalLaneGap * HEIGHT / BASE_HEIGHT)
    HORIZONTAL_LANE_START_Y = round(sim_cfg.road.horizontalLaneStartY * HEIGHT / BASE_HEIGHT)
    TREES_WIDTH = round(sim_cfg.env_objects.TREES_WIDTH * WIDTH / BASE_WIDTH)
    TREES_GAP = round(sim_cfg.env_objects.TREES_GAP * HEIGHT / BASE_HEIGHT)
    TREES_START_POINT = round(sim_cfg.env_objects.TREES_START_POINT * HEIGHT / BASE_HEIGHT)
    HOUSES_WIDTH = round(sim_cfg.env_objects.house_width * WIDTH / BASE_WIDTH)
    HOUSES_GAP = round(sim_cfg.env_objects.house_gap * HEIGHT / BASE_HEIGHT)
    HOUSE_START_POINT = round(sim_cfg.env_objects.HOUSE_START_POINT * HEIGHT / BASE_HEIGHT)
    SIDEWALK_OBJECTS_WIDTH = round(sim_cfg.env_objects.sidewalk_object_width * WIDTH / BASE_WIDTH)
    SIDEWALK_OBJECTS_GAP_MIN = round(sim_cfg.env_objects.sidewalk_object_gap_min * HEIGHT / BASE_HEIGHT)
    SIDEWALK_OBJECTS_GAP_MAX = round(sim_cfg.env_objects.sidewalk_object_gap_max * HEIGHT / BASE_HEIGHT)
    SIDEWALK_OBJECTS_START_POINT = round(sim_cfg.env_objects.sidewalk_object_start_point * HEIGHT / BASE_HEIGHT)
    TURNING_RADIUS_MULTIPLIER = round(100 * HEIGHT / BASE_HEIGHT)
